{ Ansteuerung einer LCD-Anzeige (z.B. Typ 162) mit 1 - 4 Zeilen, 16 oder 24 Zeichen

  J. Rathlev, Dec. 2010
  }  

unit Display;

{$XDATA}
{ $C-}

interface

type
  TDisplayType = (dt2x16,dt4x20);
  TCharArray = array[0..8] of byte;
  TDispString = string[40];
  
const
{ user defined characters
    xxTab[0] = 8 * ASCII-Code
    xxTab[1..8] = 5 Pixels per line }
  AuTab :  TCharArray = (8,4,$E,$15,4,4,4,4,0);   // Arrow up (Char 0)
  AdTab :  TCharArray = (16,4,4,4,4,$15,$E,4,0);  // Arrow down (Char 1) 

procedure InitDisplay (DispType : TDisplayType; Cursor,Blink : boolean);   

procedure SetDispPos (Line,CPos : byte);
procedure SetDispLine (Line : byte);
procedure ClearDispLine (Line : byte);
procedure MoveDispCursor (Left : boolean);
procedure MoveDisplay (Left : boolean);

procedure DefineDispChar (CharTab : TCharArray);
procedure WriteDispChar (AChar : char);
procedure WriteDispFixString (AText : CodePChar);
procedure WriteDispString (var AText : TDispString XData);
procedure WriteDispDec (n : byte);

var
  ToDisplay : Text; 
  
implementation

uses Math, Utils;

type
  TDisplayProps = record
    CharsInLine : byte;
    MemAddrs    : array [0..3] of byte;
    end;
    
const
  LcdRS   = %10100000;   // Signal RS
  LcdRW   = %01100000;   // Signal RW
  LcdE    = %00110000;   // Signal E
  FourBit = %0010;       // Function 4-Bit-Mode

  DisplayPropsList : array[TDisplayType] of TDisplayProps =
    ((CharsInLine : 16; MemAddrs : (0,$40,$10,$50)),
     (CharsInLine : 20; MemAddrs : (0,$40,$14,$54)));

var
  LcdPort : byte absolute P1;
  RSBit   : boolean absolute LcdPort.7;
  RWBit   : boolean absolute LcdPort.6;
  EBit    : boolean absolute LcdPort.4;
  DisplayProps : TDisplayProps;
  
// Write control byte to display
procedure WriteCtrl (CtrlCode : byte); Assembler;
Asm
  MOV     A,CtrlCode
  SWAP    A               // first Bit 4 .. 7
  ANL     A,#0FH
  ORL     A,#LcdE
  MOV     LcdPort,#LcdE   // E = 1
  MOV     LcdPort,A       // 4 Bit out
  NOP
  CLR     EBit            // E = 0
  MOV     A,CtrlCode
  ANL     A,#0FH          // Bit 0 .. 3
  ORL     A,#LcdE
  SETB    EBit            // E = 1
  MOV     LcdPort,A       // 4 Bit out
  NOP
  CLR     EBit            // E = 0
  end;

procedure WriteDispChar (AChar : char); Assembler;
Asm 
@WrChar:
    LCALL   ShortDelay
    MOV     A,AChar
    MOV     LcdPort,#LcdRS    // Data mode     
    SWAP    A                 // first Bit 4 .. 7
    ANL     A,#0FH
    ORL     A,#LcdE or LcdRS
    SETB    EBit    
    MOV     LcdPort,A         // write 4 Bit
    NOP
    CLR     EBit 
    MOV     A,AChar
    ANL     A,#0FH            // Bit 0 .. 3
    ORL     A,#LcdE or LcdRS
    SETB    EBit
    MOV     LcdPort,A         // write 4 Bit
    NOP
    CLR     EBit  
    NOP
    CLR     RSBit         
  end;
  
procedure WriteToDisplay; Assembler;
Asm
    MOV     WriteDispChar.AChar,A
    LCALL   WriteDispChar
  end;
       
// Write user defined character to display 
procedure DefineDispChar (CharTab : TCharArray);
var 
  i,n : byte;
begin
  n:=$40+CharTab[0];   // CG RAM Address Set + 8 * ASCII-Code
  for i:=1 to 8 do begin
    WriteCtrl(n+i);    
    WriteDispChar(char(CharTab[i]));
    end;
  end;
  
procedure SetDispPos (Line,CPos : byte);
begin
  Dec(CPos); Dec(Line);
  with DisplayProps do if CPos<CharsInLine then begin 
    ShortDelay;
    WriteCtrl($80+MemAddrs[Line and 3]+CPos);
    ShortDelay;
    WriteCtrl($06);  // auto increment
    end;
  end;
    
procedure SetDispLine (Line : byte);
begin
  SetDispPos(Line,1);
  end;  
   
procedure ClearDispLine (Line : byte);
var
  i : byte;
begin
  SetDispLine(Line);
  for i:=1 to DisplayProps.CharsInLine do WriteDispChar(' ');
  SetDispLine(Line);
  end;

procedure MoveDispCursor (Left : boolean);
begin
  if Left then WriteCtrl(%10000)
  else WriteCtrl(%10100); 
  end;
  
procedure MoveDisplay (Left : boolean);
begin
  if Left then WriteCtrl(%11000)
  else WriteCtrl(%11100); 
  end;
   
procedure DeleteDispChar;
begin
  MoveDispCursor(true);
  WriteDispChar(' ');
  MoveDispCursor(true);
  end;

procedure WriteDispFixString (AText : CodePChar);
var
  i : byte;
begin
//  for i:=1 to word(AText[0]) do WriteDispChar(AText[i]);
  i:=0;
  while AText[i]>#0 do begin
    WriteDispChar(AText[i]);
    inc(i);
    end;
  end;
         
procedure WriteDispString (var AText : TDispString XData);
var
  i : byte;
begin
  for i:=1 to length(AText) do WriteDispChar(AText[i]);
  end;
        
procedure WriteDispDec (n : byte);
var
  q,r : byte;
begin
  ByteDivMod(n,100,q,r);
  WriteDispChar(char(q+$30));
  ByteDivMod(r,10,q,r);
  WriteDispChar(char(q+$30));
  WriteDispChar(char(r+$30));  
  end;
   
procedure InitDisplay (DispType : TDisplayType; Cursor,Blink : boolean);
var
  ctrl : byte;   
begin
  DisplayProps:=DisplayPropsList[DispType];
  LcdPort:=0;
  LcdPort:=LcdE;
  LcdPort:=LcdE or FourBit;       // set to 4 bit mode
  ShortDelay;
  EBit:=false;
  LongDelay(5);

  WriteCtrl(FourBit shl 4 or 8);  // 2 or 4 lines, 5x7
  ShortDelay; ShortDelay;
  ctrl:=$C;
  if Cursor then ctrl:=ctrl or 2;
  if Blink then ctrl:=ctrl or 1;
  WriteCtrl(ctrl);                  // display on, cursor on, cursor blink
  LongDelay(5);

  WriteCtrl(1);                   // clear display, cursor home
  LongDelay(5);
  
  WriteCtrl(6);                   // auto increment cursor
  LongDelay(5);
  Assign (ToDisplay,WriteToDisplay);   
  end;
  
end.